/*

*/

import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

/**
   Implement our own "thread pool". A "thread pool" is
   two objects, an array of worker threads and a single
   blocking queue (that is shared by the worker threads).

   The blocking queue holds Runnable objects that will
   be called(!) by the worker threads. The key idea is
   that the Runnable objects are not used to create threads
   (the worker thread already exists), the run() method in
   each Runnable object is called by a worker thread. So the
   worker thread executes the Runnable's run method (to
   completion!) and then the worker thread gets another
   Runnable object, from the blocking queue, and calls its
   run() method until the worker thread gets interrupted.

   Having one blocking queue for all the worker threads (as
   opposed to having a blocking queue for each worker thread)
   has a disadvantage and an advantage. The disadvantage is
   that the worker threads may end up blocking each other when
   they go to the shared blocking queue to get a Runnable object.
   If there are many worker threads and they are working on very
   short lived tasks, then the worker threads will be constantly
   asking the shared queue for new Runnable objects. If many
   worker threads simultaneously request a new Runnable from the
   single blocking queue, then all but one of them must wait (this
   is called "lock contention"). The advantage of having a single
   blocking queue is that we get good load balancing. As long as
   there are items in the blocking queue, no worker thread will be
   idle.
*/
public class MyThreadPool_v1
{
   private final int poolSize;
   private final Thread[] threadPool;
   private final BlockingQueue<Runnable> taskQueue;

   public MyThreadPool_v1(int poolSize)
   {
      this.poolSize = poolSize;
      threadPool = new Thread[poolSize];
      taskQueue = new LinkedBlockingQueue<>();

      // Create, and start, the worker threads.
      for (int i = 0; i < poolSize; ++i)
      {
         threadPool[i] = new WorkerThread(taskQueue);
         threadPool[i].start(); // Start a worker, which will block on its empty queue.
      }
   }


   /**
      Submit a new task to the task queue.

      This blocks if the task queue is full or if
      some other thread is currently adding a task
      to the queue.
   */
   public void execute(Runnable task)
   {
      taskQueue.add(task);
   }


   /**
      Send every worker thread a interrupt message
      and then wait for each worker thread to stop.
   */
   public void shutdown()
   {
      System.out.println("***** Shutting down the thread pool.");
      try
      {
         for (int i = 0; i < poolSize; ++i)
         {
            threadPool[i].interrupt();
         }
         for (int i = 0; i < poolSize; ++i)
         {
            threadPool[i].join();
         }
      }
      catch (InterruptedException e)
      {
      }
   }


   /**

   */
   boolean awaitTermination(long timeout, java.util.concurrent.TimeUnit unit)
   throws InterruptedException
   {
      System.out.println("***** Waiting for the thread pool to terminate.");
      try
      {
         for (int i = 0; i < poolSize; ++i)
         {
            threadPool[i].join();
         }
      }
      catch (InterruptedException e)
      {
      }
      return true;
   }


   /**
      This inner class implements the worker threads that
      are in the "thread pool". The worker threads share a
      queue of Runnable tasks that they should execute.
      A worker thread executes the run() method in each
      Runnable task that it removes from the task queue.

      Notice how the Runnable objects are NOT used to create
      threads! The Runnable objects are simply convenient
      containers for the code that makes up a task.
   */
   class WorkerThread extends Thread
   {
      private final BlockingQueue<Runnable> taskQueue;

      public WorkerThread(BlockingQueue<Runnable> taskQueue)
      {
         this.taskQueue = taskQueue; // Shared task queue.
      }

      @Override
      public void run()
      {
         boolean done = false;
         System.out.println("***** Worker thread " + this.getId() + " has started up.");
         while ( (! done) || (! taskQueue.isEmpty()) )
         {
            try
            {
               // Get a Runnable from the task queue.
               final Runnable task = taskQueue.take();
               // Execute the Runnable's run() method in this thread.
               task.run();

               // Check if the thread pool is supposed to shut down.
               if ( Thread.interrupted() )
               {
                  done = true;
               }
            }
            catch (InterruptedException e)
            {
               done = true;
               //System.out.println("***** Worker thread  " + this.getId() + ", InterruptedException.");
            }
         }
         System.out.println("***** Worker thread " + this.getId() + " has shut down.");
      }
   }
}
